from os import path
from autoconf import conf
import autofit as af
from test_autofit.integration.src.dataset import dataset as ds
from test_autofit.integration.src.phase import phase as ph
from test_autofit.integration.src.model import profiles
from test_autofit.integration.tests import runner

import numpy as np

test_type = "searches"
data_name = "gaussian_x1"


class GridPhase(af.as_grid_search(phase_class=ph.Phase, parallel=False)):
    @property
    def grid_priors(self):
        return [self.model.profiles.gaussian.one_tuple]


phase = GridPhase(
    name="phase_grid",
    profiles=af.CollectionPriorModel(gaussian=profiles.Gaussian),
    search=af.DynestyStatic(n_live_points=40, evidence_tolerance=5.0),
    number_of_steps=2,
)

integration_path = path.join("{}".format(path.dirname(path.realpath(__file__))), "..", "..")

conf.instance.push(
    new_path=path.join(integration_path, "config"),
    output_path=path.join(integration_path, "output", test_type)
)

dataset = ds.Dataset.from_fits(
    data_path=path.join(integration_path, "dataset", data_name, "data.fits"),
    noise_map_path=path.join(integration_path, "dataset", data_name, "noise_map.fits"),
)

phase.run(dataset=dataset)



import autofit as af
from test_autofit.integration.src.phase import phase as ph
from test_autofit.integration.src.model import profiles
from test_autofit.integration.tests import runner

test_type = "searches"
data_name = "gaussian_x1"

phase = ph.Phase(
    name="phase",
    profiles=af.CollectionPriorModel(gaussian=profiles.Gaussian),
    search=af.Emcee(),
)

if __name__ == "__main__":
    import sys

    runner.run(module=sys.modules[__name__])



import os
from os import path
import pickle
import shutil

import numpy as np
import pytest

import autofit as af
from autoconf import conf
from autofit.mock import mock
from autofit.mock.mock_search import MockSamples

directory = path.dirname(path.realpath(__file__))
pytestmark = pytest.mark.filterwarnings("ignore::FutureWarning")


@pytest.fixture(autouse=True)
def set_config_path():
    conf.instance.push(
        new_path=path.join(directory, "files", "nlo", "config"),
        output_path=path.join(directory, "files", "nlo", "output"),
    )


@pytest.fixture(name="mapper")
def make_mapper():
    return af.ModelMapper()


@pytest.fixture(name="mock_list")
def make_mock_list():
    return [af.PriorModel(mock.MockClassx4), af.PriorModel(mock.MockClassx4)]


@pytest.fixture(name="result")
def make_result():
    mapper = af.ModelMapper()
    mapper.component = mock.MockClassx2Tuple
    # noinspection PyTypeChecker
    return af.Result(
        samples=MockSamples(gaussian_tuples=[(0, 0), (1, 0)]),
        previous_model=mapper,
        search=mock.MockSearch(),
    )


class TestResult:
    def test_model(self, result):
        component = result.model.component
        assert component.one_tuple.one_tuple_0.mean == 0
        assert component.one_tuple.one_tuple_1.mean == 1
        assert component.one_tuple.one_tuple_0.sigma == 0.2
        assert component.one_tuple.one_tuple_1.sigma == 0.2

    def test_model_absolute(self, result):
        component = result.model_absolute(a=2.0).component
        assert component.one_tuple.one_tuple_0.mean == 0
        assert component.one_tuple.one_tuple_1.mean == 1
        assert component.one_tuple.one_tuple_0.sigma == 2.0
        assert component.one_tuple.one_tuple_1.sigma == 2.0

    def test_model_relative(self, result):
        component = result.model_relative(r=1.0).component
        assert component.one_tuple.one_tuple_0.mean == 0
        assert component.one_tuple.one_tuple_1.mean == 1
        assert component.one_tuple.one_tuple_0.sigma == 0.0
        assert component.one_tuple.one_tuple_1.sigma == 1.0

    def test_raises(self, result):
        with pytest.raises(af.exc.PriorException):
            result.model.mapper_from_gaussian_tuples(
                result.samples.gaussian_tuples, a=2.0, r=1.0
            )


class TestCopyWithNameExtension:
    @staticmethod
    def assert_non_linear_attributes_equal(copy):
        assert copy.paths.name ==  path.join("name", "one")

    def test_copy_with_name_extension(self):
        search = af.MockSearch(af.Paths("name", tag="tag"))
        copy = search.copy_with_name_extension("one")

        self.assert_non_linear_attributes_equal(copy)
        assert search.paths.tag == copy.paths.tag


@pytest.fixture(name="nlo_setup_path")
def test_nlo_setup():
    nlo_setup_path = path.join("{}".format(path.dirname(path.realpath(__file__))), "files", "nlo", "setup")

    if path.exists(nlo_setup_path):
        shutil.rmtree(nlo_setup_path)

    os.mkdir(nlo_setup_path)

    return nlo_setup_path


@pytest.fixture(name="nlo_model_info_path")
def test_nlo_model_info():
    nlo_model_info_path = path.join(
        "{}".format(path.dirname(path.realpath(__file__))), "file", "nlo", "model_info"
    )

    if path.exists(nlo_model_info_path):
        shutil.rmtree(nlo_model_info_path)

    return nlo_model_info_path


@pytest.fixture(name="nlo_wrong_info_path")
def test_nlo_wrong_info():
    nlo_wrong_info_path = path.join("{}".format(
        path.dirname(path.realpath(__file__))
    ), "files", "nlo", "wrong_info")

    if path.exists(nlo_wrong_info_path):
        shutil.rmtree(nlo_wrong_info_path)

    os.mkdir(nlo_wrong_info_path)

    return nlo_wrong_info_path


class TestLabels:
    def test_param_names(self):
        model = af.PriorModel(mock.MockClassx4)
        assert [
                   "one",
                   "two",
                   "three",
                   "four",
               ] == model.model_component_and_parameter_names

    def test_label_config(self):
        assert conf.instance["notation"]["label"]["label"]["one"] == "one_label"
        assert conf.instance["notation"]["label"]["label"]["two"] == "two_label"
        assert conf.instance["notation"]["label"]["label"]["three"] == "three_label"
        assert conf.instance["notation"]["label"]["label"]["four"] == "four_label"


test_path = path.join(
    "{}".format(path.dirname(path.realpath(__file__))), "files", "phase"
)


class TestMovePickleFiles:
    def test__move_pickle_files(self):

        output_path = path.join(
            "{}".format(path.dirname(path.realpath(__file__))),
            "files",
            "nlo",
            "output",
            "test_phase",
            "mock",
            "pickles",
        )

        if path.exists(output_path):
            shutil.rmtree(output_path)

        search = af.MockSearch(paths=af.Paths(name="test_phase"))

        pickle_paths = [
            path.join(
                "{}".format(path.dirname(path.realpath(__file__))), "files", "pickles"
            )
        ]

        arr = np.ones((3, 3))

        with open(path.join(pickle_paths[0], "test.pickle"), "wb") as f:
            pickle.dump(arr, f)

        pickle_paths = [
            path.join(
                "{}".format(path.dirname(path.realpath(__file__))),
                "files",
                "pickles",
                "test.pickle",
            )
        ]

        search.move_pickle_files(pickle_files=pickle_paths)

        with open(path.join(output_path, "test.pickle"), "rb") as f:
            arr_load = pickle.load(f)

        assert (arr == arr_load).all()

        if path.exists(test_path):
            shutil.rmtree(test_path)



import pytest

import autofit as af
from autofit.mock import mock


@pytest.fixture(name="phase")
def make_phase():
    return af.AbstractPhase(
        search=mock.MockSearch(paths=af.Paths(name="name", tag="phase_tag"))
    )


def test_metadata_dictionary(phase):
    phase.pipeline_name = "pipeline_name"
    phase.pipeline_tag = "pipeline_tag"
    assert phase._default_metadata == {
        "phase": "name",
        "phase_tag": "phase_tag",
        "pipeline": "pipeline_name",
        "pipeline_tag": "pipeline_tag",
    }



from copy import copy
from copy import copy
from itertools import count
from os import path
from typing import List, Generator, Callable, Type, Union, Tuple

from autofit import AbstractPriorModel, ModelInstance, Paths, Result, Analysis, NonLinearSearch
from autofit.non_linear.grid.grid_search import make_lists
from autofit.non_linear.parallel import AbstractJob, Process, AbstractJobResult


class JobResult(AbstractJobResult):
    def __init__(
            self,
            number: int,
            result: Result,
            perturbed_result: Result
    ):
        """
        The result of a single sensitivity comparison

        Parameters
        ----------
        result
        perturbed_result
        """
        super().__init__(number)
        self.result = result
        self.perturbed_result = perturbed_result

    @property
    def log_likelihood_difference(self):
        return self.perturbed_result.log_likelihood - self.result.log_likelihood


class Job(AbstractJob):
    _number = count()

    def __init__(
            self,
            analysis: Analysis,
            model: AbstractPriorModel,
            perturbation_model: AbstractPriorModel,
            search: NonLinearSearch
    ):
        """
        Job to run non-linear searches comparing how well a model and a model with a perturbation
        fit the image.

        Parameters
        ----------
        model
            A base model that fits the image without a perturbation
        perturbation_model
            A model of the perturbation which has been added to the underlying image
        analysis
            A class definition which can compares instances of a model to a perturbed image
        search
            A non-linear search
        """
        super().__init__()
        self.analysis = analysis
        self.model = model

        self.perturbation_model = perturbation_model

        paths = search.paths

        self.search = search
        self.perturbed_search = search.copy_with_paths(
            Paths(
                name=paths.name,
                tag=paths.tag + "[perturbed]",
                path_prefix=paths.path_prefix,
                remove_files=paths.remove_files,
            )
        )

    def perform(self) -> JobResult:
        """
        - Create one model with a perturbation and another without
        - Fit each model against the perturbed image

        Returns
        -------
        An object comprising the results of the two fits
        """
        result = self.search.fit(
            model=self.model,
            analysis=self.analysis
        )

        perturbed_model = copy(self.model)
        perturbed_model.perturbation = self.perturbation_model

        perturbed_result = self.perturbed_search.fit(
            model=perturbed_model,
            analysis=self.analysis
        )
        return JobResult(
            number=self.number,
            result=result,
            perturbed_result=perturbed_result
        )


class SensitivityResult:
    def __init__(self, results: List[JobResult]):
        self.results = sorted(results)

    def __getitem__(self, item):
        return self.results[item]

    def __iter__(self):
        return iter(self.results)

    def __len__(self):
        return len(self.results)


class Sensitivity:
    def __init__(
            self,
            base_instance,
            base_model: AbstractPriorModel,
            perturbation_model: AbstractPriorModel,
            simulate_function: Callable,
            analysis_class: Type[Analysis],
            search: NonLinearSearch,
            step_size: Union[Tuple[float], float] = 0.1,
            number_of_cores: int = 2
    ):
        """
        Perform sensitivity mapping to evaluate whether a perturbation
        can be detected if it occurs in different parts of an image.

        For a range from 0 to 1 with step_size, for each dimension of the
        perturbation_model, a perturbation is created and used in conjunction
        with the instance to create an image.

        For each of these images, a fit is run with just the model and with both
        the model and perturbation_model to compare how much better the image
        can be fit if the perturbation is included.

        Parameters
        ----------
        base_instance
            An instance of a model to which perturbations are applied prior to
            images being generated
        base_model
            A model that fits the instance well
        search
            A NonLinear search class which is copied and used to evaluate fitness
        analysis_class
            A class which can compare an image to an instance and evaluate fitness
        perturbation_model
            A model which provides a perturbations to be applied to the instance
            before creating images
        simulate_function
            A function that can convert an instance into an image
        step_size
            The size of the step between perturbations. For example, a set size of 0.5
            with a perturbation_model of dimension 3 would give (1 / 0.5) ^ 3 = 8
            distinct perturbations.
        number_of_cores
            How many cores does this computer have? Minimum 2.
        """
        self.instance = base_instance
        self.model = base_model

        self.search = search
        self.analysis_class = analysis_class

        self.step_size = step_size
        self.perturbation_model = perturbation_model
        self.simulate_function = simulate_function
        self.number_of_cores = number_of_cores

    def run(self) -> SensitivityResult:
        """
        Run fits and comparisons for all perturbations, returning
        a list of results.
        """
        results = list()
        for result in Process.run_jobs(
                self._make_jobs(),
                number_of_cores=self.number_of_cores
        ):
            results.append(result)
        return SensitivityResult(results)

    @property
    def _lists(self) -> List[List[float]]:
        """
        A list of hypercube vectors, used to instantiate
        the perturbation_model and create the individual
        perturbations.
        """
        return make_lists(
            self.perturbation_model.prior_count,
            step_size=self.step_size
        )

    @property
    def _labels(self) -> Generator[str, None, None]:
        """
        One label for each perturbation, used to distinguish
        fits for each perturbation by placing them in separate
        directories.
        """
        for list_ in self._lists:
            strings = list()
            for value, prior_tuple in zip(
                    list_,
                    self.perturbation_model.prior_tuples
            ):
                path, prior = prior_tuple
                value = prior.value_for(
                    value
                )
                strings.append(
                    f"{path}_{value}"
                )
            yield "_".join(strings)

    @property
    def _perturbation_instances(self) -> Generator[
        ModelInstance, None, None
    ]:
        """
        A list of instances each of which defines a perturbation to
        be applied to the image.
        """
        for list_ in self._lists:
            yield self.perturbation_model.instance_from_unit_vector(
                list_
            )

    @property
    def _searches(self) -> Generator[
        NonLinearSearch, None, None
    ]:
        """
        A list of non-linear searches, each of which is applied to
        one perturbation.
        """
        for label in self._labels:
            paths = self.search.paths
            name_path = path.join(
                paths.name,
                paths.tag,
                paths.non_linear_tag,
                label,
            )
            yield self._search_instance(
                name_path
            )

    def _search_instance(
            self,
            name_path: str
    ) -> NonLinearSearch:
        """
        Create a search instance, distinguished by its name

        Parameters
        ----------
        name_path
            A path to distinguish this search from other searches

        Returns
        -------
        A non linear search, copied from the instance search
        """
        paths = self.search.paths
        search_instance = self.search.copy_with_paths(
            Paths(
                name=name_path,
                tag=paths.tag,
                path_prefix=paths.path_prefix,
                remove_files=paths.remove_files,
            )
        )

        return search_instance

    def _make_jobs(self) -> Generator[Job, None, None]:
        """
        Create a list of jobs to be run on separate processes.

        Each job fits a perturbed image with the original model
        and a model which includes a perturbation.
        """
        for perturbation_instance, search in zip(
                self._perturbation_instances,
                self._searches
        ):
            instance = copy(self.instance)
            instance.perturbation = perturbation_instance
            dataset = self.simulate_function(
                instance
            )
            yield Job(
                analysis=self.analysis_class(
                    dataset
                ),
                model=self.model,
                perturbation_model=self.perturbation_model,
                search=search
            )



import autofit as af
from test_autofit.integration.src.dataset.dataset import Dataset
from test_autofit.integration.src.phase.result import Result
from test_autofit.integration.src.phase.analysis import Analysis
from test_autofit.integration.src.phase.meta_dataset import MetaDataset
from test_autofit.integration.src.phase.settings import SettingsPhase

import numpy as np

# The 'phase.py' module is mostly unchanged from the previous tutorial, however the 'run' function has been updated.


class Phase(af.AbstractPhase):

    profiles = af.PhaseProperty("profiles")

    Result = Result

    @af.convert_paths
    def __init__(self, paths, *, profiles, settings=SettingsPhase(), search=af.Emcee):
        """
        A phase which fits a model composed of multiple line profiles (Gaussian, Exponential) using a non-linear search.

        Parameters
        ----------
        paths : af.Paths
            Handles the output directory structure.
        profiles : [profiles.Profile]
            The model components (e.g. Gaussian, Exponential) fitted by this phase.
        search: class
            The class of a non_linear search
        data_trim_left : int or None
            The number of pixels by which the data is trimmed from the left-hand side.
        data_trim_right : int or None
            The number of pixels by which the data is trimmed from the right-hand side.
        """

        paths.tag = settings.tag

        super().__init__(paths=paths, search=search)

        self.profiles = profiles

        self.meta_dataset = MetaDataset(settings=settings)

    @property
    def folders(self):
        return self.search.path_prefix

    def run(self, dataset: Dataset, info=None, results=None):
        """
        Pass a dataset to the phase, running the phase and non-linear search.

        Parameters
        ----------
        dataset: aa.Dataset
            The dataset fitted by the phase, as defined in the 'dataset.py' module.
        mask: Mask2D
            The mask used for the analysis.

        Returns
        -------
        result: AbstractPhase.Result
            A result object comprising information on the `NonLinearSearch` and the maximum likelihood model.
        """

        mask = np.full(fill_value=False, shape=dataset.data.shape)

        # These functions save the objects we will later access using the aggregator. They are saved via the 'pickle'
        # module in Python, which serializes the data on to the hard-disk.

        # See the 'dataset.py' module for a description of what the metadata is.

        self.save_metadata(dataset=dataset)
        self.save_dataset(dataset=dataset)
        self.save_mask(mask=mask)
        self.save_meta_dataset(meta_dataset=self.meta_dataset)

        self.model = self.model.populate(results)

        results = results or af.ResultsCollection()

        # This saves the search information of the phase, meaning that we can use the search instance
        # (e.g. Emcee) to interpret our results in the aggregator.

        analysis = self.make_analysis(dataset=dataset, mask=mask)

        result = self.run_analysis(analysis=analysis, info=info)

        return self.make_result(result=result, analysis=analysis)

    def make_analysis(self, dataset, mask):
        """
        Returns an Analysis object, which creates the dataset and contains the functions which perform the fit.

        Parameters
        ----------
        dataset: aa.Dataset
            The dataset fitted by the phase, as defined in the 'dataset.py' module.

        Returns
        -------
        analysis : Analysis
            An analysis object that the `NonLinearSearch` calls to determine the fit log_likelihood for a given model
            instance.
        """

        masked_dataset = self.meta_dataset.masked_dataset_from_dataset_and_mask(
            dataset=dataset, mask=mask
        )

        return Analysis(
            masked_dataset=masked_dataset, image_path=self.search.paths.visualize_path
        )

    def make_result(self, result, analysis):
        return self.Result(samples=result.samples, analysis=analysis)



from os import path
import pickle
import sys

import numpy as np
import pytest

import autofit as af
from autoconf import conf
from autofit.mock import mock

directory = path.dirname(path.realpath(__file__))
pytestmark = pytest.mark.filterwarnings("ignore::FutureWarning")


@pytest.fixture(autouse=True)
def set_config_path():
    conf.instance.push(
        new_path=path.join(directory, "files", "dynesty", "config"),
        output_path=path.join(directory, "files", "dynesty", "output"),
    )


class MockDynestyResults:
    def __init__(self, samples, logl, logwt, ncall, logz, nlive):
        self.samples = samples
        self.logl = logl
        self.logwt = logwt
        self.ncall = ncall
        self.logz = logz
        self.nlive = nlive


class MockDynestySampler:
    def __init__(self, results):
        self.results = results


class TestDynestyConfig:
    def test__loads_from_config_file_if_not_input(self):
        dynesty = af.DynestyStatic(
            prior_passer=af.PriorPasser(sigma=2.0, use_errors=False, use_widths=False),
            n_live_points=151,
            facc=0.6,
            evidence_tolerance=0.1,
            bound="ellipse",
            sample="manual",
            update_interval=True,
            bootstrap=1.0,
            enlarge=2.0,
            vol_dec=2.1,
            vol_check=2.2,
            walks=26,
            slices=6,
            fmove=0.8,
            max_move=101,
            maxiter=2,
            maxcall=3,
            logl_max=1.0,
            n_effective=4,
            iterations_per_update=501,
            terminate_at_acceptance_ratio=False,
            acceptance_ratio_threshold=0.5,
            number_of_cores=2,
        )

        assert dynesty.prior_passer.sigma == 2.0
        assert dynesty.prior_passer.use_errors == False
        assert dynesty.prior_passer.use_widths == False
        assert dynesty.iterations_per_update == 501
        assert dynesty.n_live_points == 151
        assert dynesty.facc == 0.6
        assert dynesty.evidence_tolerance == 0.1
        assert dynesty.bound == "ellipse"
        assert dynesty.sample == "manual"
        assert dynesty.update_interval == True
        assert dynesty.bootstrap == 1.0
        assert dynesty.enlarge == 2.0
        assert dynesty.vol_dec == 2.1
        assert dynesty.vol_check == 2.2
        assert dynesty.walks == 26
        assert dynesty.slices == 6
        assert dynesty.fmove == 0.8
        assert dynesty.max_move == 101
        assert dynesty.maxiter == 2
        assert dynesty.maxcall == 3
        assert dynesty.logl_max == 1.0
        assert dynesty.n_effective == 4
        assert dynesty.terminate_at_acceptance_ratio == False
        assert dynesty.acceptance_ratio_threshold == 0.5
        assert dynesty.number_of_cores == 2

        dynesty = af.DynestyStatic()

        assert dynesty.prior_passer.sigma == 3.0
        assert dynesty.prior_passer.use_errors == True
        assert dynesty.prior_passer.use_widths == True
        assert dynesty.iterations_per_update == 500
        assert dynesty.n_live_points == 150
        assert dynesty.evidence_tolerance == 0.159
        assert dynesty.bound == "multi"
        assert dynesty.sample == "auto"
        assert dynesty.update_interval == None
        assert dynesty.bootstrap == 0.0
        assert dynesty.enlarge == 1.0
        assert dynesty.vol_dec == 0.5
        assert dynesty.vol_check == 2.0
        assert dynesty.walks == 5
        assert dynesty.facc == 0.5
        assert dynesty.slices == 5
        assert dynesty.fmove == 0.9
        assert dynesty.max_move == 100
        assert dynesty.maxiter == sys.maxsize
        assert dynesty.maxcall == sys.maxsize
        assert dynesty.logl_max == np.inf
        assert dynesty.n_effective == np.inf
        assert dynesty.terminate_at_acceptance_ratio == True
        assert dynesty.acceptance_ratio_threshold == 2.0
        assert dynesty.number_of_cores == 1

        dynesty = af.DynestyDynamic(
            prior_passer=af.PriorPasser(sigma=2.0, use_errors=False, use_widths=False),
            n_live_points=0,
            iterations_per_update=501,
            facc=0.6,
            evidence_tolerance=0.2,
            bound="ellipse",
            sample="manual",
            update_interval=True,
            bootstrap=1.0,
            enlarge=2.0,
            vol_dec=2.1,
            vol_check=2.2,
            walks=26,
            slices=6,
            fmove=0.8,
            max_move=101,
            maxiter=2,
            maxcall=3,
            logl_max=1.0,
            n_effective=4,
            terminate_at_acceptance_ratio=False,
            acceptance_ratio_threshold=0.5,
            number_of_cores=3,
        )

        assert dynesty.prior_passer.sigma == 2.0
        assert dynesty.prior_passer.use_errors == False
        assert dynesty.prior_passer.use_widths == False
        assert dynesty.n_live_points == 500
        assert dynesty.iterations_per_update == 501
        assert dynesty.facc == 0.6
        assert dynesty.evidence_tolerance == 0.2
        assert dynesty.bound == "ellipse"
        assert dynesty.sample == "manual"
        assert dynesty.update_interval == True
        assert dynesty.bootstrap == 1.0
        assert dynesty.enlarge == 2.0
        assert dynesty.vol_dec == 2.1
        assert dynesty.vol_check == 2.2
        assert dynesty.walks == 26
        assert dynesty.slices == 6
        assert dynesty.fmove == 0.8
        assert dynesty.max_move == 101
        assert dynesty.maxiter == 2
        assert dynesty.maxcall == 3
        assert dynesty.logl_max == 1.0
        assert dynesty.n_effective == 4
        assert dynesty.terminate_at_acceptance_ratio == False
        assert dynesty.acceptance_ratio_threshold == 0.5
        assert dynesty.number_of_cores == 3

        dynesty = af.DynestyDynamic()

        assert dynesty.prior_passer.sigma == 3.0
        assert dynesty.prior_passer.use_errors == True
        assert dynesty.prior_passer.use_widths == True
        assert dynesty.n_live_points == 5
        assert dynesty.iterations_per_update == 501
        assert dynesty.facc == 0.6
        assert dynesty.bound == "balls"
        assert dynesty.sample == "rwalk"
        assert dynesty.update_interval == 2.0
        assert dynesty.bootstrap == 1.0
        assert dynesty.enlarge == 2.0
        assert dynesty.vol_dec == 0.4
        assert dynesty.vol_check == 3.0
        assert dynesty.walks == 26
        assert dynesty.slices == 6
        assert dynesty.fmove == 0.8
        assert dynesty.max_move == 101
        assert dynesty.maxiter == sys.maxsize
        assert dynesty.maxcall == sys.maxsize
        assert dynesty.logl_max == np.inf
        assert dynesty.n_effective == np.inf
        assert dynesty.terminate_at_acceptance_ratio == True
        assert dynesty.acceptance_ratio_threshold == 2.0
        assert dynesty.number_of_cores == 4

    def test__tag(self):
        dynesty = af.DynestyStatic(
            n_live_points=40, bound="none", sample="auto", enlarge=1.0
        )

        assert (
            dynesty.tag
            == "dynesty_static[nlive_40__bound_none__enlarge_1.0__sample_auto]"
        )

        dynesty = af.DynestyStatic(
            n_live_points=41,
            bound="multi",
            sample="unif",
            enlarge=1.0,
            vol_dec=2.0,
            vol_check=3.0,
        )

        assert (
            dynesty.tag
            == "dynesty_static[nlive_41__bound_multi_vol_dec_2.0_vol_check_3.0__enlarge_1.0__sample_unif]"
        )

        dynesty = af.DynestyStatic(
            n_live_points=43,
            bound="single",
            sample="rwalk",
            enlarge=1.0,
            walks=1,
            facc=0.5,
        )

        assert (
            dynesty.tag
            == "dynesty_static[nlive_43__bound_single__enlarge_1.0__sample_rwalk_walks_1_facc_0.5]"
        )

        dynesty = af.DynestyStatic(
            n_live_points=44,
            bound="balls",
            sample="hslice",
            enlarge=1.0,
            slices=1,
            max_move=2,
        )

        assert (
            dynesty.tag
            == "dynesty_static[nlive_44__bound_balls__enlarge_1.0__sample_hslice_slices_1_max_move_2]"
        )

        dynesty = af.DynestyStatic(
            n_live_points=44, bound="balls", sample="slice", enlarge=1.0, slices=1
        )

        assert (
            dynesty.tag
            == "dynesty_static[nlive_44__bound_balls__enlarge_1.0__sample_slice_slices_1]"
        )

        dynesty = af.DynestyStatic(
            n_live_points=44, bound="balls", sample="rslice", enlarge=1.0, slices=1
        )

        assert (
            dynesty.tag
            == "dynesty_static[nlive_44__bound_balls__enlarge_1.0__sample_rslice_slices_1]"
        )

        dynesty = af.DynestyDynamic(
            bound="multi", sample="unif", enlarge=1.0, vol_dec=2.0, vol_check=3.0
        )

        assert (
            dynesty.tag
            == "dynesty_dynamic[nlive_5__bound_multi_vol_dec_2.0_vol_check_3.0__enlarge_1.0__sample_unif]"
        )

        dynesty = af.DynestyDynamic(
            bound="balls", sample="rslice", enlarge=1.0, slices=1
        )

        assert (
            dynesty.tag
            == "dynesty_dynamic[nlive_5__bound_balls__enlarge_1.0__sample_rslice_slices_1]"
        )

    def test__samples_from_model(self):
        # Setup pickle of mock Dynesty sampler that the samples_from_model function uses.

        results = MockDynestyResults(
            samples=np.array(
                [[1.0, 2.0, 3.0, 5.0], [1.0, 2.0, 3.0, 4.0], [1.0, 2.0, 3.0, 4.0]]
            ),
            logl=[1.0, 2.0, 3.0],
            logwt=[np.log(1.0), np.log(2.0), np.log(3.0)],
            ncall=[5.0, 5.0],
            logz=[-2.0, -1.0, 0.0],
            nlive=3,
        )

        sampler = MockDynestySampler(results=results)

        paths = af.Paths()

        dynesty = af.DynestyStatic(paths=paths)

        with open(path.join(dynesty.paths.samples_path, "dynesty.pickle"), "wb") as f:
            pickle.dump(sampler, f)

        model = af.ModelMapper(mock_class=mock.MockClassx4)
        model.mock_class.two = af.LogUniformPrior(lower_limit=0.0, upper_limit=10.0)

        samples = dynesty.samples_via_sampler_from_model(model=model)

        assert isinstance(samples.parameters, list)
        assert isinstance(samples.parameters[0], list)
        assert isinstance(samples.log_likelihoods, list)
        assert isinstance(samples.log_priors, list)
        assert isinstance(samples.log_posteriors, list)
        assert isinstance(samples.weights, list)

        assert samples.parameters == [
            [1.0, 2.0, 3.0, 5.0],
            [1.0, 2.0, 3.0, 4.0],
            [1.0, 2.0, 3.0, 4.0],
        ]
        assert samples.log_likelihoods == [1.0, 2.0, 3.0]
        assert samples.log_priors == [0.2, 0.25, 0.25]
        assert samples.weights == pytest.approx([1.0, 2.0, 3.0], 1.0e-4)
        assert samples.total_samples == 10
        assert samples.log_evidence == 0.0
        assert samples.number_live_points == 3


class TestCopyWithNameExtension:
    @staticmethod
    def assert_non_linear_attributes_equal(copy):
        assert copy.paths.name == path.join("name", "one")

    def test_dynesty(self):
        search = af.DynestyStatic(af.Paths("name"))

        copy = search.copy_with_name_extension("one")
        self.assert_non_linear_attributes_equal(copy)
        assert isinstance(copy, af.DynestyStatic)
        assert copy.prior_passer is search.prior_passer
        assert (
            copy.terminate_at_acceptance_ratio is search.terminate_at_acceptance_ratio
        )
        assert copy.acceptance_ratio_threshold is search.acceptance_ratio_threshold

        assert copy.iterations_per_update is search.iterations_per_update
        assert copy.evidence_tolerance == search.evidence_tolerance
        assert copy.n_live_points == search.n_live_points
        assert copy.bound == search.bound
        assert copy.sample == search.sample
        assert copy.update_interval == search.update_interval
        assert copy.bootstrap == search.bootstrap
        assert copy.enlarge == search.enlarge
        assert copy.vol_dec == search.vol_dec
        assert copy.vol_check == search.vol_check
        assert copy.walks == search.walks
        assert copy.facc == search.facc
        assert copy.slices == search.slices
        assert copy.fmove == search.fmove
        assert copy.max_move == search.max_move
        assert copy.number_of_cores == search.number_of_cores

        search = af.DynestyDynamic(af.Paths("name"))

        copy = search.copy_with_name_extension("one")
        self.assert_non_linear_attributes_equal(copy)
        assert isinstance(copy, af.DynestyDynamic)
        assert copy.prior_passer is search.prior_passer
        assert (
            copy.terminate_at_acceptance_ratio is search.terminate_at_acceptance_ratio
        )
        assert copy.acceptance_ratio_threshold is search.acceptance_ratio_threshold

        assert copy.evidence_tolerance == search.evidence_tolerance
        assert copy.iterations_per_update is search.iterations_per_update
        assert copy.bound == search.bound
        assert copy.sample == search.sample
        assert copy.update_interval == search.update_interval
        assert copy.bootstrap == search.bootstrap
        assert copy.enlarge == search.enlarge
        assert copy.vol_dec == search.vol_dec
        assert copy.vol_check == search.vol_check
        assert copy.walks == search.walks
        assert copy.facc == search.facc
        assert copy.slices == search.slices
        assert copy.fmove == search.fmove
        assert copy.max_move == search.max_move
        assert copy.number_of_cores == search.number_of_cores



import json

from autoconf import conf
from autofit.non_linear import samples as samp
from autofit.non_linear.abstract_search import NonLinearSearch


class AbstractMCMC(NonLinearSearch):
    @property
    def config_type(self):
        return conf.instance["non_linear"]["mcmc"]

    def samples_via_csv_json_from_model(self, model):
        parameters, log_likelihoods, log_priors, log_posteriors, weights = samp.load_from_table(
            filename=self.paths.samples_file, model=model
        )

        with open(self.paths.info_file) as infile:
            samples_info = json.load(infile)

        return samp.MCMCSamples(
            model=model,
            parameters=parameters,
            log_likelihoods=log_likelihoods,
            log_priors=log_priors,
            weights=weights,
            auto_correlation_times=samples_info["auto_correlation_times"],
            auto_correlation_check_size=samples_info["auto_correlation_check_size"],
            auto_correlation_required_length=samples_info[
                "auto_correlation_required_length"
            ],
            auto_correlation_change_threshold=samples_info[
                "auto_correlation_change_threshold"
            ],
            total_walkers=samples_info["total_walkers"],
            total_steps=samples_info["total_steps"],
            time=samples_info["time"],
        )



from autofit.tools.add_notebook_quotes import add_notebook_quotes


def test():
    lines = [
        '"""',
        "docs",
        '"""',
        "code",
        '"""',
        "docs",
        '"""'
    ]

    assert add_notebook_quotes(
        lines
    ) == [
               '%%',
               '"""',
               "docs",
               '"""',
               '%%',
               "code",
               '%%',
               '"""',
               "docs",
               '"""',
               '%%'
           ]



from os import path

import autofit as af
from autoconf import conf
from autofit.mock.mock_search import MockSearch, MockAnalysis, MockSamples
from autofit.mock import mock

directory = path.dirname(path.realpath(__file__))

conf.instance = conf.Config(path.join(directory, "files", "config"))


def make_pipeline_1(name):
    search = MockSearch(
        name="phase_1", samples=MockSamples(gaussian_tuples=[(0.5, 0.5)])
    )
    phase = af.Phase(
        model=af.PriorModel(mock.MockComponents, parameter=af.GaussianPrior(10.0, 1.0)),
        search=search,
        analysis_class=MockAnalysis,
    )
    return af.Pipeline(f"{name}_1", "", None, phase)


def make_pipeline_2(name):
    search = MockSearch(
        name="phase_2", samples=MockSamples(gaussian_tuples=[(0.5, 0.5)])
    )
    phase = af.Phase(
        model=af.PriorModel(mock.MockComponents, parameter=af.last.model.parameter),
        search=search,
        analysis_class=MockAnalysis,
    )
    return af.Pipeline(f"{name}_2", "", None, phase)


def make_pipeline(name,):
    pipeline_2 = make_pipeline_2(name)
    pipeline_1 = make_pipeline_1(name)

    return pipeline_1 + pipeline_2


def test_pipeline_composition():
    pipeline = make_pipeline("test")
    results = pipeline.run(mock.MockDataset())
    assert results[0].model.parameter.mean == results[1].model.parameter.mean



from os import path
import pytest

from autoconf import conf
import autofit as af
from autofit.mock import mock

directory = path.dirname(path.realpath(__file__))
pytestmark = pytest.mark.filterwarnings("ignore::FutureWarning")


@pytest.fixture(autouse=True)
def set_config_path():
    conf.instance.push(
        new_path=path.join(directory, "files", "pyswarms", "config"),
        output_path=path.join(directory, "files", "pyswarms", "output"),
    )


class TestPySwarmsGlobalConfig:
    def test__loads_from_config_file_correct(self):

        pso = af.PySwarmsGlobal(
            prior_passer=af.PriorPasser(sigma=2.0, use_errors=False, use_widths=False),
            n_particles=51,
            iters=2001,
            cognitive=0.4,
            social=0.5,
            inertia=0.6,
            initializer=af.InitializerBall(lower_limit=0.2, upper_limit=0.8),
            iterations_per_update=10,
            number_of_cores=2,
        )

        assert pso.prior_passer.sigma == 2.0
        assert pso.prior_passer.use_errors == False
        assert pso.prior_passer.use_widths == False
        assert pso.n_particles == 51
        assert pso.iters == 2001
        assert pso.cognitive == 0.4
        assert pso.social == 0.5
        assert pso.inertia == 0.6
        assert isinstance(pso.initializer, af.InitializerBall)
        assert pso.initializer.lower_limit == 0.2
        assert pso.initializer.upper_limit == 0.8
        assert pso.iterations_per_update == 10
        assert pso.number_of_cores == 2

        pso = af.PySwarmsGlobal()

        assert pso.prior_passer.sigma == 3.0
        assert pso.prior_passer.use_errors == True
        assert pso.prior_passer.use_widths == True
        assert pso.n_particles == 50
        assert pso.iters == 2000
        assert pso.cognitive == 0.1
        assert pso.social == 0.2
        assert pso.inertia == 0.3
        assert isinstance(pso.initializer, af.InitializerPrior)
        assert pso.iterations_per_update == 11
        assert pso.number_of_cores == 1

        pso = af.PySwarmsLocal(
            prior_passer=af.PriorPasser(sigma=2.0, use_errors=False, use_widths=False),
            n_particles=51,
            iters=2001,
            cognitive=0.4,
            social=0.5,
            inertia=0.6,
            number_of_k_neighbors=4,
            minkowski_p_norm=1,
            initializer=af.InitializerBall(lower_limit=0.2, upper_limit=0.8),
            iterations_per_update=10,
            number_of_cores=2,
        )

        assert pso.prior_passer.sigma == 2.0
        assert pso.prior_passer.use_errors == False
        assert pso.prior_passer.use_widths == False
        assert pso.n_particles == 51
        assert pso.iters == 2001
        assert pso.cognitive == 0.4
        assert pso.social == 0.5
        assert pso.inertia == 0.6
        assert pso.number_of_k_neighbors == 4
        assert pso.minkowski_p_norm == 1
        assert isinstance(pso.initializer, af.InitializerBall)
        assert pso.initializer.lower_limit == 0.2
        assert pso.initializer.upper_limit == 0.8
        assert pso.iterations_per_update == 10
        assert pso.number_of_cores == 2

        pso = af.PySwarmsLocal()

        assert pso.prior_passer.sigma == 3.0
        assert pso.prior_passer.use_errors == True
        assert pso.prior_passer.use_widths == True
        assert pso.n_particles == 50
        assert pso.iters == 2000
        assert pso.cognitive == 0.1
        assert pso.social == 0.2
        assert pso.inertia == 0.3
        assert pso.number_of_k_neighbors == 3
        assert pso.minkowski_p_norm == 2
        assert isinstance(pso.initializer, af.InitializerPrior)
        assert pso.iterations_per_update == 11
        assert pso.number_of_cores == 1

    def test__tag(self):

        pso = af.PySwarmsGlobal(
            n_particles=51, iters=2001, cognitive=0.4, social=0.5, inertia=0.6
        )

        assert pso.tag == "pyswarms_global[particles_51_c_0.4_s_0.5_i_0.6]"

        pso = af.PySwarmsLocal(
            n_particles=51, iters=2001, cognitive=0.4, social=0.5, inertia=0.6
        )

        assert pso.tag == "pyswarms_local[particles_51_c_0.4_s_0.5_i_0.6]"

    def test__samples_from_model(self):

        pyswarms = af.PySwarmsGlobal(paths=af.Paths())

        model = af.ModelMapper(mock_class=mock.MockClassx3)
        model.mock_class.one = af.LogUniformPrior(lower_limit=0.0, upper_limit=100.0)
        model.mock_class.two = af.LogUniformPrior(lower_limit=0.0, upper_limit=100.0)
        model.mock_class.three = af.LogUniformPrior(lower_limit=0.0, upper_limit=100.0)
        # model.mock_class.four = af.LogUniformPrior(lower_limit=0.0, upper_limit=100.0)

        samples = pyswarms.samples_via_sampler_from_model(model=model)

        assert isinstance(samples.parameters, list)
        assert isinstance(samples.parameters[0], list)
        assert isinstance(samples.log_likelihoods, list)
        assert isinstance(samples.log_priors, list)
        assert isinstance(samples.log_posteriors, list)

        assert samples.parameters[0] == pytest.approx(
            [50.1254, 1.04626, 10.09456], 1.0e-4
        )

        assert samples.log_likelihoods[0] == pytest.approx(-5071.80777, 1.0e-4)
        assert samples.log_posteriors[0] == pytest.approx(-5070.73298, 1.0e-4)
        assert samples.weights[0] == 1.0

        assert len(samples.parameters) == 500
        assert len(samples.log_likelihoods) == 500


class TestCopyWithNameExtension:
    @staticmethod
    def assert_non_linear_attributes_equal(copy):
        assert copy.paths.name == path.join("name", "one")

    def test__pyswarms(self):
        search = af.PySwarmsGlobal(af.Paths("name"))

        copy = search.copy_with_name_extension("one")
        self.assert_non_linear_attributes_equal(copy)
        assert isinstance(copy, af.PySwarmsGlobal)
        assert copy.prior_passer is search.prior_passer
        assert copy.n_particles is search.n_particles
        assert copy.iters is search.iters
        assert copy.cognitive == search.cognitive
        assert copy.social == search.social
        assert copy.inertia == search.inertia
        assert copy.ftol is search.ftol
        assert copy.initializer is search.initializer
        assert copy.iterations_per_update is search.iterations_per_update
        assert copy.number_of_cores is search.number_of_cores



import os
from os import path
from autoconf import conf
from test_autofit.integration.src.dataset import dataset as ds


def run(module):

    integration_path = path.join("{}".format(os.path.dirname(os.path.realpath(__file__))), "..")

    conf.instance.push(
        new_path=path.join(integration_path,"config"),
        output_path=path.join(integration_path, "output", module.test_type),
    )

    dataset = ds.Dataset.from_fits(
        data_path=path.join(integration_path, "dataset",module.data_name, "data.fits"),
        noise_map_path=path.join(integration_path,"dataset",module.data_name, "noise_map.fits"),
    )

    module.phase.run(dataset=dataset)



import autofit as af
from autofit.mock import mock


class TestMixin:
    def test_mixin(self):
        class MyPhase(af.as_grid_search(af.AbstractPhase)):
            Result = mock.MockResult

            @property
            def grid_priors(self):
                return [self.model.component.one_tuple.one_tuple_0]

            def run(self):
                analysis = mock.MockAnalysis()
                return self.run_analysis(analysis)

        my_phase = MyPhase(
            number_of_steps=2,
            search=mock.MockSearch(
                samples=mock.MockSamples(gaussian_tuples=[(0.5, 0.5), (0.5, 0.5)])
            ),
        )
        my_phase.model.component = mock.MockClassx2Tuple

        result = my_phase.run()

        assert isinstance(result, af.GridSearchResult)
        assert len(result.results) == 2
        assert len(result.lower_limit_lists) == 2
        assert len(result.physical_lower_limits_lists) == 2

        assert isinstance(result.best_result, af.Result)

    def test_parallel_flag(self):
        my_phase = af.as_grid_search(af.AbstractPhase, parallel=True)(
            search=mock.MockSearch()
        )
        assert my_phase.search.parallel



from os import path
import shutil

import pytest

import autofit as af
from autoconf import conf
from autofit.mock import mock

directory = path.dirname(path.realpath(__file__))
pytestmark = pytest.mark.filterwarnings("ignore::FutureWarning")


@pytest.fixture(autouse=True)
def set_config_path():
    conf.instance.push(
        new_path=path.join(directory, "files", "emcee", "config"),
        output_path=path.join(directory, "files", "emcee", "output"),
    )


class TestEmceeConfig:
    def test__loads_from_config_file_correct(self):
        emcee = af.Emcee(
            prior_passer=af.PriorPasser(sigma=2.0, use_errors=False, use_widths=False),
            nwalkers=51,
            nsteps=2001,
            initializer=af.InitializerBall(lower_limit=0.2, upper_limit=0.8),
            auto_correlation_check_for_convergence=False,
            auto_correlation_check_size=101,
            auto_correlation_required_length=51,
            auto_correlation_change_threshold=0.02,
            number_of_cores=2,
        )

        assert emcee.prior_passer.sigma == 2.0
        assert emcee.prior_passer.use_errors == False
        assert emcee.prior_passer.use_widths == False
        assert emcee.nwalkers == 51
        assert emcee.nsteps == 2001
        assert isinstance(emcee.initializer, af.InitializerBall)
        assert emcee.initializer.lower_limit == 0.2
        assert emcee.initializer.upper_limit == 0.8
        assert emcee.auto_correlation_check_for_convergence == False
        assert emcee.auto_correlation_check_size == 101
        assert emcee.auto_correlation_required_length == 51
        assert emcee.auto_correlation_change_threshold == 0.02
        assert emcee.number_of_cores == 2

        emcee = af.Emcee()

        assert emcee.prior_passer.sigma == 3.0
        assert emcee.prior_passer.use_errors == True
        assert emcee.prior_passer.use_widths == True
        assert emcee.nwalkers == 50
        assert emcee.nsteps == 2000
        assert isinstance(emcee.initializer, af.InitializerPrior)
        assert emcee.auto_correlation_check_for_convergence == True
        assert emcee.auto_correlation_check_size == 100
        assert emcee.auto_correlation_required_length == 50
        assert emcee.auto_correlation_change_threshold == 0.01
        assert emcee.number_of_cores == 1

    def test__tag(self):
        emcee = af.Emcee(nwalkers=11)

        assert emcee.tag == "emcee[nwalkers_11]"

    def test__samples_from_model(self):
        emcee = af.Emcee(paths=af.Paths())

        shutil.copy(
            path.join(directory, "files", "emcee.hdf"),
            path.join(
                directory, "files", "emcee", "output", "emcee[nwalkers_50]", "samples"
            ),
        )

        model = af.ModelMapper(mock_class=mock.MockClassx4)
        model.mock_class.two = af.LogUniformPrior(lower_limit=0.0, upper_limit=10.0)

        samples = emcee.samples_via_sampler_from_model(model=model)

        assert isinstance(samples.parameters, list)
        assert isinstance(samples.parameters[0], list)
        assert isinstance(samples.log_likelihoods, list)
        assert isinstance(samples.log_priors, list)
        assert isinstance(samples.log_posteriors, list)
        assert isinstance(samples.weights, list)

        assert samples.parameters[0] == pytest.approx(
            [0.173670, 0.162607, 3095.28, 0.62104], 1.0e-4
        )
        assert samples.log_likelihoods[0] == pytest.approx(-17257775239.32677, 1.0e-4)
        assert samples.log_priors[0] == pytest.approx(1.6102016075510708, 1.0e-4)
        assert samples.weights[0] == pytest.approx(1.0, 1.0e-4)
        assert samples.total_steps == 1000
        assert samples.total_walkers == 10
        assert samples.auto_correlation_times[0] == pytest.approx(31.98507, 1.0e-4)


class TestEmceeOutput:
    def test__median_pdf_parameters(self):
        emcee = af.Emcee(paths=af.Paths())
        shutil.copy(
            path.join(directory, "files", "emcee.hdf"),
            path.join(
                directory, "files", "emcee", "output", "emcee[nwalkers_50]", "samples"
            ),
        )

        model = af.ModelMapper(mock_class=mock.MockClassx4)
        model.mock_class.two = af.LogUniformPrior(lower_limit=0.0, upper_limit=10.0)

        samples = emcee.samples_via_sampler_from_model(model=model)

        assert samples.median_pdf_vector == pytest.approx(
            [0.008422, -0.026413, 9.9579656, 0.494618], 1.0e-3
        )

    def test__vector_at_sigma__uses_output_files(self):
        emcee = af.Emcee(paths=af.Paths())

        shutil.copy(
            path.join(directory, "files", "emcee.hdf"),
            path.join(
                directory, "files", "emcee", "output", "emcee[nwalkers_50]", "samples"
            ),
        )

        model = af.ModelMapper(mock_class=mock.MockClassx4)
        model.mock_class.two = af.LogUniformPrior(lower_limit=0.0, upper_limit=10.0)

        samples = emcee.samples_via_sampler_from_model(model=model)

        parameters = samples.vector_at_sigma(sigma=3.0)

        assert parameters[0][0:2] == pytest.approx((-0.003197, 0.019923), 1e-2)

        parameters = samples.vector_at_sigma(sigma=1.0)

        assert parameters[0][0:2] == pytest.approx((0.0042278, 0.01087681), 1e-2)

    def test__autocorrelation_times(self):
        emcee = af.Emcee(paths=af.Paths())

        shutil.copy(
            path.join(directory, "files", "emcee.hdf"),
            path.join(
                directory, "files", "emcee", "output", "emcee[nwalkers_50]", "samples"
            ),
        )

        model = af.ModelMapper(mock_class=mock.MockClassx4)
        model.mock_class.two = af.LogUniformPrior(lower_limit=0.0, upper_limit=10.0)

        samples = emcee.samples_via_sampler_from_model(model=model)

        assert samples.previous_auto_correlation_times == pytest.approx(
            [31.1079, 36.0910, 72.44768, 65.86194], 1.0e-4
        )
        assert samples.auto_correlation_times == pytest.approx(
            [31.98507, 36.51001, 73.47629, 67.67495], 1.0e-4
        )


class TestCopyWithNameExtension:
    @staticmethod
    def assert_non_linear_attributes_equal(copy):
        assert copy.paths.name == path.join("name", "one")

    def test_emcee(self):
        search = af.Emcee(af.Paths("name"))

        copy = search.copy_with_name_extension("one")
        self.assert_non_linear_attributes_equal(copy)
        assert isinstance(copy, af.Emcee)
        assert copy.prior_passer is search.prior_passer
        assert copy.nwalkers is search.nwalkers
        assert copy.nsteps is search.nsteps
        assert copy.initializer is search.initializer
        assert (
            copy.auto_correlation_check_for_convergence
            is search.auto_correlation_check_for_convergence
        )
        assert copy.auto_correlation_check_size is search.auto_correlation_check_size
        assert (
            copy.auto_correlation_required_length
            is search.auto_correlation_required_length
        )
        assert (
            copy.auto_correlation_change_threshold
            is search.auto_correlation_change_threshold
        )
        assert copy.number_of_cores is search.number_of_cores
